/* license: AGPLv3 or later */
/* copyright 2024 zamfofex */

#include <stdlib.h>
#include "chapter.h"

struct mimimi_chapter_I {
	unsigned char colors[1024 * 512];
	struct mimimi_history history;
	int progression;
	struct mimimi_toast_dialogue tutorial, dialogue;
	struct mimimi_stage stage;
};

static void mimimi_chapter_I_background(struct mimimi_chapter_I *chapter)
{
	int i;
	struct mimimi_area *area;
	struct mimimi_image *image;
	
	for (i = 0 ; i < chapter->stage.area_count ; i++) {
		area = chapter->stage.areas + i;
		area->background_count = 1;
		area->backgrounds = malloc(sizeof *area->backgrounds);
		if (area->backgrounds == NULL) exit(1);
		mimimi_background(&area->backgrounds[0].image, area->ground);
		area->backgrounds[0].x = 0;
		area->backgrounds[0].y = 0;
		area->backgrounds[0].z = 0;
	}
	
	area = chapter->stage.areas;
	
	image = &area->backgrounds[0].image;
	free(image->colors);
	image->width = mimimi_chapter_I_background_width;
	image->height = mimimi_chapter_I_background_height;
	image->colors = malloc(image->width * image->height);
	if (image->colors == NULL) exit(1);
	mimimi_unpack_bits(image->width * image->height, image->colors, mimimi_chapter_I_background_packed);
	area->backgrounds[0].x = -88;
	area->backgrounds[0].y = 0;
	area->backgrounds[0].z = 0;
	
	area->foregrounds = malloc(sizeof *area->foregrounds);
	if (area->foregrounds == NULL) exit(1);
	
	image = &area->foregrounds[0].image;
	image->width = mimimi_chapter_I_foreground_width;
	image->height = mimimi_chapter_I_foreground_height;
	image->colors = malloc(image->width * image->height);
	if (image->colors == NULL) exit(1);
	mimimi_unpack_bits(image->width * image->height, image->colors, mimimi_chapter_I_foreground_packed);
	area->foreground_count = 1;
	area->foregrounds[0].x = 0;
	area->foregrounds[0].y = 0;
	area->foregrounds[0].z = 0;
}

static void mimimi_chapter_I_size(struct mimimi_image *image)
{
	int width, height;
	
	if (image->width / image->height < 2) {
		height = 256;
		width = height * image->width / image->height;
	}
	else {
		width = 512;
		height = width * image->height / image->width;
	}
	
	if (width < 256) {
		width = 256;
		height = width * image->height / image->width;
		if (height > 384) height = 384;
	}
	if (height < 192) {
		height = 192;
		width = height * image->width / image->height;
		if (width > 512) width = 512;
	}
	
	image->width = width;
	image->height = height;
}

static void mimimi_chapter_I_tick(void *data, struct mimimi_input *input)
{
	struct mimimi_chapter_I *chapter;
	struct mimimi_sprite *player;
	struct mimimi_stage_jump jump;
	
	chapter = data;
	player = chapter->stage.sprites;
	
	mimimi_chapter_I_size(&input->image);
	input->image.colors = chapter->colors;
	
	jump.stage = &chapter->stage;
	jump.sprite = player;
	
	mimimi_history_tick(&chapter->history, input->left, input->right);
	mimimi_controls_tick(&player->walk, &chapter->history, &mimimi_stage_jump, &jump);
	mimimi_stage_tick(&chapter->stage, input);
	
	if (chapter->progression == 0) {
		if (chapter->stage.area_index == 1) chapter->progression++;
	}
	if (chapter->progression == 1) {
		if (mimimi_toast_dialogue_tick(&chapter->tutorial, &input->image) != 0) chapter->progression++;
	}
	if (chapter->progression == 2) {
		if (chapter->stage.area_index == chapter->stage.area_count - 1) chapter->progression++;
	}
	if (chapter->progression == 3) {
		if (mimimi_toast_dialogue_tick(&chapter->dialogue, &input->image) != 0) chapter->progression++;
	}
	if (chapter->progression == 4) {
		mimimi_play_sound(&chapter->stage, &mimimi_chapter_I_music);
		chapter->progression++;
	}
}

int main(void)
{
	static struct mimimi_ground ground0 = {4, 3, {0, 0x6B}, {0}, {0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0}};
	static struct mimimi_ground ground1 = {5, 3, {0, 8}, {0}, {0, 0, 0, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0, 0}};
	static struct mimimi_dialogue_paragraph paragraphs[] = {{NULL, "Mirai", "There seems to be nothing for me to do. Hmmm... Maybe I can find something to do later."}, {NULL, "Ryouna", "Hear that? If you come back later, then maybe you can find something to do!"}, {NULL, "Ryoubi", "Who are you talking to? Are you stupid or something?"}, {NULL, "Ryouna", "Aaah! Ryoubi, I really like when you yell insults at me!"}, {NULL, "Mirai", "Wha-? What are you two doing here?"}, {NULL, "Haruka", "Hehe, don't worry about it!"}};
	static struct mimimi_dialogue_paragraph tutorial = {NULL, "Haruka", "Press a direction twice to jump!"};
	
	struct mimimi_dialogue dialogue;
	struct mimimi_chapter_I *chapter;
	struct mimimi_platform_set *platforms;
	struct mimimi_image *image;
	int count;
	
	tutorial.model = mimimi_haruka_flash;
	paragraphs[0].model = mimimi_mirai_flash;
	paragraphs[1].model = mimimi_ryouna_flash;
	paragraphs[2].model = mimimi_ryoubi_flash;
	paragraphs[3].model = mimimi_ryouna_flash;
	paragraphs[4].model = mimimi_mirai_flash;
	paragraphs[5].model = mimimi_haruka_flash;
	
	chapter = malloc(sizeof *chapter);
	if (chapter == NULL) exit(1);
	
	count = 0;
	while (mimimi_chapter_I_stage[count]->width != 0) count++;
	
	mimimi_stage(&chapter->stage, mimimi_chapter_I_stage, count);
	mimimi_chapter_I_background(chapter);
	
	chapter->progression = 0;
	
	dialogue.count = 1;
	dialogue.paragraphs = &tutorial;
	mimimi_toast_dialogue(&chapter->tutorial, &dialogue);
	
	dialogue.count = sizeof paragraphs / sizeof *paragraphs;
	dialogue.paragraphs = paragraphs;
	mimimi_toast_dialogue(&chapter->dialogue, &dialogue);
	
	mimimi_spawn(&chapter->stage, mimimi_mirai_flash, 0, 0, 80, 256);
	
	image = malloc(sizeof *image);
	if (image == NULL) return 1;
	
	mimimi_background(image, &ground0);
	
	platforms = mimimi_add_platform_set(&chapter->stage);
	platforms->angular_weight = 512;
	platforms->strength = 128;
	
	mimimi_spawn_angular_platform(&chapter->stage, platforms, &ground0, image, -0x5800, -0x1C80, -80, 256, 1024);
	mimimi_spawn_angular_platform(&chapter->stage, platforms, &ground0, image, -0x5800, -0x1C80, -80, 256, 2048);
	mimimi_spawn_angular_platform(&chapter->stage, platforms, &ground0, image, -0x5800, -0x1C80, -80, 256, 3072);
	
	platforms = mimimi_add_platform_set(&chapter->stage);
	platforms->angular_weight = 512;
	platforms->strength = 128;
	
	mimimi_spawn_angular_platform(&chapter->stage, platforms, &ground0, image, -0x6500, -0x2100, 80, 256, 1024);
	mimimi_spawn_angular_platform(&chapter->stage, platforms, &ground0, image, -0x6500, -0x2100, 80, 256, 2048);
	mimimi_spawn_angular_platform(&chapter->stage, platforms, &ground0, image, -0x6500, -0x2100, 80, 256, 3072);
	
	image = malloc(sizeof *image);
	if (image == NULL) return 1;
	
	mimimi_background(image, &ground1);
	
	platforms = mimimi_add_platform_set(&chapter->stage);
	platforms->strength = 0;
	platforms->min = 0;
	platforms->max = 1024;
	
	mimimi_spawn_linear_platform(&chapter->stage, platforms, &ground1, image, -0x4B00, -0x700, 256);
	mimimi_spawn_linear_platform(&chapter->stage, platforms, &ground1, image, -0x4800, -0x800, -256);
	mimimi_spawn_linear_platform(&chapter->stage, platforms, &ground1, image, -0x4F00, -0xC00, 256);
	mimimi_spawn_linear_platform(&chapter->stage, platforms, &ground1, image, -0x4F00, -0xD00, -256);
	
	chapter->history.left = 0;
	chapter->history.right = 0;
	
	mimimi_main(&mimimi_chapter_I_tick, chapter);
	return 1;
}
