/* license: AGPLv3 or later */
/* copyright 2024 zamfofex */

#include "../mimimi.h"

void mimimi_camera_tick(struct mimimi_position *camera, struct mimimi_position *position, int ox, int oy)
{
	camera->x *= 3;
	camera->y *= 3;
	camera->x += position->x + ox;
	camera->y += position->y + oy;
	camera->x /= 4;
	camera->y /= 4;
}

void mimimi_clamped_camera_tick(struct mimimi_clamped_camera *camera, int width0, int height0)
{
	struct mimimi_position *input;
	struct mimimi_position *output;
	int x, y, width, height;
	int left, right;
	int top, bottom;
	
	input = camera->input;
	output = &camera->output;
	
	x = camera->offset.x;
	y = camera->offset.y;
	width = camera->size.width * 128;
	height = camera->size.height * 128;
	
	left = x * 128;
	right = left + width;
	
	top = y * 128;
	bottom = top + height;
	
	width0 *= 8;
	height0 *= 8;
	
	left += width0 / 2;
	right -= width0 / 2;
	top += height0 / 2;
	bottom -= height0 / 2;
	
	*output = *input;
	
	if (output->x < left) output->x = left;
	if (output->x > right) output->x = right;
	if (output->y < top) output->y = top;
	if (output->y > bottom) output->y = bottom;
	
	if (width < width0) output->x = left - (width0 - width) / 2;
	if (height < height0) output->y = top - (height0 - height) / 2;
}

void mimimi_boxing(struct mimimi_size *size, struct mimimi_image *image)
{
	int width, height;
	int x, y;
	
	width = image->width - size->width * 16;
	height = image->height - size->height * 16;
	
	width = mimimi_div_up(width, 2);
	height = mimimi_div_up(height, 2);
	
	for (y = 0 ; y <= height ; y++) {
		for (x = 0 ; x < image->width ; x++) {
			image->colors[x + image->width * y] = 1;
			image->colors[x + image->width * (image->height - y)] = 1;
		}
	}
	for (y = 0 ; y < image->height ; y++) {
		for (x = 0 ; x <= width ; x++) {
			image->colors[x + image->width * y] = 1;
			image->colors[(image->width - x) + image->width * y] = 1;
		}
	}
}
