/* license: AGPLv3 or later */
/* copyright 2024 zamfofex */

#include <stdlib.h>
#include "../mimimi.h"

static void mimimi_box(struct mimimi_image *image, unsigned char stroke, unsigned char fill)
{
	static unsigned char colors[10][10] = {{0, 0, 0, 0, 0, 0, 0, 0, 0, 1}, {0, 0, 0, 0, 0, 0, 0, 1, 1, 1}, {0, 0, 0, 0, 0, 1, 1, 1, 1, 1}, {0, 0, 0, 0, 1, 1, 1, 1, 2, 2}, {0, 0, 0, 1, 1, 1, 2, 2, 2, 2}, {0, 0, 1, 1, 1, 2, 2, 2, 2, 2}, {0, 0, 1, 1, 2, 2, 2, 2, 2, 2}, {0, 1, 1, 1, 2, 2, 2, 2, 2, 2}, {0, 1, 1, 2, 2, 2, 2, 2, 2, 2}, {1, 1, 1, 2, 2, 2, 2, 2, 2, 2}, };
	
	int width, height;
	int x0, y0, x1, y1;
	unsigned char table[3];
	unsigned char color;
	
	width = image->width;
	height = image->height;
	
	table[0] = 0;
	table[1] = stroke;
	table[2] = fill;
	
	for (y1 = 0 ; y1 < height ; y1++) {
		for (x1 = 0 ; x1 < width ; x1++) {
			image->colors[x1 + y1 * width] = fill;
		}
	}
	for (x1 = 0 ; x1 < width ; x1++) {
		image->colors[x1 + 0 * width] = stroke;
		image->colors[x1 + 1 * width] = stroke;
		image->colors[x1 + (height - 1) * width] = stroke;
		image->colors[x1 + (height - 2) * width] = stroke;
	}
	
	for (y1 = 0 ; y1 < height ; y1++) {
		image->colors[0 + y1 * width] = stroke;
		image->colors[1 + y1 * width] = stroke;
		image->colors[(width - 1) + y1 * width] = stroke;
		image->colors[(width - 2) + y1 * width] = stroke;
	}
	
	for (y0 = 0 ; y0 < 10 ; y0++) {
		
		for (x0 = 0 ; x0 < 10 ; x0++) {
			
			x1 = width - 1 - x0;
			y1 = height - 1 - y0;
			
			color = table[colors[y0][x0]];
			
			image->colors[x0 + y0 * width] = color;
			image->colors[x1 + y0 * width] = color;
			image->colors[x0 + y1 * width] = color;
			image->colors[x1 + y1 * width] = color;
		}
	}
}

static void mimimi_portrait(struct mimimi_image *image, struct mimimi_model *model, unsigned char background)
{
	static unsigned char mask[6][6] = {{0, 0, 0, 0, 0, 0}, {0, 0, 0, 0, 1, 1}, {0, 0, 0, 1, 1, 1}, {0, 0, 1, 1, 1, 1}, {0, 1, 1, 1, 1, 1}, {0, 1, 1, 1, 1, 1}, };
	
	int width, height;
	struct mimimi_pose pose = {0};
	int i;
	int x0, y0, x1, y1;
	
	width = image->width;
	height = image->height;
	
	pose.layers[0].y_angle = 42;
	
	pose.x = width / 2;
	pose.y = height / 2 + 16;
	
	for (i = 0 ; i < width * height ; i++) image->colors[i] = background;
	
	mimimi_pose(image, model, &pose);
	
	for (y0 = 0 ; y0 < 6 ; y0++) {
		
		for (x0 = 0 ; x0 < 6 ; x0++) {
			
			if (mask[y0][x0] != 0) continue;
			
			x1 = width - 1 - x0;
			y1 = height - 1 - y0;
			
			image->colors[x0 + y0 * width] = 0;
			image->colors[x1 + y0 * width] = 0;
			image->colors[x0 + y1 * width] = 0;
			image->colors[x1 + y1 * width] = 0;
		}
	}
}

static void mimimi_dialogue_box(struct mimimi_image *box, struct mimimi_model *model, char *name, int width, char *paragraph)
{
	char *text;
	int count;
	int height;
	unsigned char colors[18 * 18];
	struct mimimi_image portrait;
	int n, i;
	
	text = paragraph;
	count = 0;
	while (*text != 0) {
		text = mimimi_skip_line(mimimi_font, text, width - 12);
		count++;
	}
	
	height = 30 + 14 * count;
	
	box->width = width;
	box->height = height;
	mimimi_box(box, 4, 13);
	
	portrait.width = 18;
	portrait.height = 18;
	portrait.colors = colors;
	mimimi_portrait(&portrait, model, 8);
	mimimi_stamp(box, 6, 6, &portrait);
	
	n = mimimi_draw_paragraph(mimimi_font, name, box, 28, 7, 4);
	for (i = 28 ; i < n - 1 ; i++) box->colors[i + 18 * box->width] = 4;
	
	mimimi_draw_text(mimimi_font, paragraph, width - 12, 14, box, 6, 24, 4);
}

void mimimi_toast_dialogue(struct mimimi_toast_dialogue *toast_dialogue, struct mimimi_dialogue *dialogue)
{
	struct mimimi_image *images;
	int i;
	struct mimimi_dialogue_paragraph *paragraph;
	int count;
	
	toast_dialogue->images = malloc(dialogue->count * sizeof *toast_dialogue->images);
	toast_dialogue->delays = malloc(dialogue->count * sizeof *toast_dialogue->delays);
	toast_dialogue->heights = malloc(dialogue->count * sizeof *toast_dialogue->heights);
	
	if (toast_dialogue->images == NULL) exit(1);
	if (toast_dialogue->delays == NULL) exit(1);
	if (toast_dialogue->heights == NULL) exit(1);
	
	images = toast_dialogue->images;
	
	for (i = 0 ; i < dialogue->count ; i++) {
		
		paragraph = dialogue->paragraphs + i;
		
		images[i].width = 128;
		images[i].height = 1024;
		
		images[i].colors = malloc(images[i].width * images[i].height);
		if (images[i].colors == NULL) exit(1);
		mimimi_dialogue_box(images + i, paragraph->model, paragraph->name, images[i].width, paragraph->text);
		toast_dialogue->heights[i] = images[i].height + 8;
		
		count = 0;
		while (paragraph->text[count] != 0) count++;
		toast_dialogue->delays[i] = count * 2;
	}
	
	toast_dialogue->count = dialogue->count;
	toast_dialogue->i = 0;
	toast_dialogue->j = 0;
	toast_dialogue->k = 0;
}

int mimimi_toast_dialogue_tick(struct mimimi_toast_dialogue *dialogue, struct mimimi_image *image)
{
	int i;
	int height;
	
	i = dialogue->i;
	if (i >= dialogue->count) return 1;
	
	height = dialogue->heights[i];
	
	if (dialogue->k == 0) {
		mimimi_stamp(image, 8, image->height - dialogue->j * height / 8, dialogue->images + i);
		dialogue->j++;
		if (dialogue->j >= 8) {
			dialogue->j = 0;
			dialogue->k++;
		}
		return 0;
	}
	
	if (dialogue->k == 1) {
		mimimi_stamp(image, 8, image->height - height, dialogue->images + i);
		dialogue->j++;
		if (dialogue->j >= dialogue->delays[i]) {
			dialogue->j = 0;
			dialogue->k++;
		}
		return 0;
	}
	
	if (dialogue->k == 2) {
		mimimi_stamp(image, 8 - (128 + 8) * dialogue->j / 16, image->height - height, dialogue->images + i);
		dialogue->j++;
		if (dialogue->j >= 16) {
			dialogue->j = 0;
			dialogue->k = 0;
			dialogue->i++;
		}
		return 0;
	}
	
	/* unreachable */
	return 0;
}
