/* license: AGPLv3 or later */
/* copyright 2024 zamfofex */
/* note: a lot of this file assumes C99 division rules */
/* note: is this worthwhile fixing? */

#include "../mimimi.h"

void mimimi_camera_stamp(struct mimimi_image *target, struct mimimi_position *camera, int x, int y, int z, struct mimimi_image *source)
{
	z += 8;
	if (z <= 0) return;
	
	x -= camera->x;
	x = mimimi_div_up(x, z);
	x += target->width / 2;
	
	y -= camera->y;
	y = mimimi_div_up(y, z);
	y += target->height / 2;
	
	mimimi_stamp(target, x, y, source);
}

static void mimimi_sprite_jumping(struct mimimi_image *image, struct mimimi_sprite *sprite)
{
	static struct mimimi_pose pose0 = {0};
	static struct mimimi_pose poses[2] = {0};
	
	int slant;
	int n;
	int dx, dy;
	int y_angle;
	
	n = sprite->direction == 1 ? 1 : -1;
	
	dx = sprite->physics.dx * -n;
	dy = sprite->physics.dy;
	
	if (dx < 0) dx = 0;
	if (dx > 8191) dx = 8191;
	
	dy += 8192;
	if (dy < 0) dy = 0;
	if (dy > 16383) dy = 16383;
	
	slant = 64 * dx / 8192 * n;
	y_angle = 64 + (20 + 20 * dx / 8192) * n;
	
	poses[0] = pose0;
	poses[1] = pose0;
	
	poses[0].layers[0].y_angle = y_angle;
	poses[0].layers[0].slant = -slant * 2;
	poses[0].layers[1].slant = slant * 2;
	poses[0].layers[4].slant = -slant / 2;
	poses[0].layers[5].slant = -slant / 2;
	
	poses[1].layers[0].y_angle = y_angle;
	poses[1].layers[2].slant = -slant * 2;
	poses[1].layers[3].slant = -slant * 2;
	poses[1].layers[4].slant = slant * 2;
	poses[1].layers[5].slant = slant * 2;
	
	poses[0].x = image->width / 2;
	poses[0].y = image->height - 15;
	poses[1].x = image->width / 2;
	poses[1].y = image->height - 15;
	
	mimimi_interpolate(poses, poses + 1, dy, 16384);
	mimimi_pose(image, sprite->model, poses);
}

static void mimimi_sprite_walking(struct mimimi_image *image, struct mimimi_sprite *sprite)
{
	static struct mimimi_pose pose0 = {0};
	static struct mimimi_pose poses[2] = {0};
	
	int slant, y_angle;
	int intent;
	int n;
	int dx;
	struct mimimi_pose *a, *b;
	int i;
	int y;
	
	n = sprite->direction == 1 ? 1 : -1;
	
	dx = sprite->physics.dx * -n;
	if (dx < 0) dx = 0;
	if (dx > 8191) dx = 8191;
	
	intent = dx / 16;
	slant = intent * n;
	y_angle = 64 + (20 + 20 * dx / 8192) * n;
	
	poses[0] = pose0;
	poses[1] = pose0;
	
	poses[0].layers[0].y_angle = y_angle;
	poses[1].layers[0].y_angle = y_angle;
	
	poses[0].layers[0].slant = -slant / 8;
	poses[1].layers[0].slant = -slant / 8;
	
	poses[0].layers[1].slant = slant / 8;
	poses[1].layers[1].slant = slant / 8;
	
	poses[1].layers[4].slant = slant / 2;
	poses[0].layers[4].slant = -slant / 3;
	poses[1].layers[8].slant = -slant / 8;
	poses[0].layers[8].slant = -slant / 16;
	
	poses[0].layers[5].slant = slant / 2;
	poses[1].layers[5].slant = -slant / 4;
	poses[0].layers[9].slant = -slant / 8;
	poses[1].layers[9].slant = -slant / 16;
	
	poses[0].layers[2].slant = slant / 6;
	poses[1].layers[2].slant = -slant / 2;
	poses[0].layers[6].slant = slant / 6;
	poses[1].layers[6].slant = slant / 2;
	poses[1].layers[6].y = intent / 128;
	
	poses[1].layers[3].slant = slant / 6;
	poses[0].layers[3].slant = -slant / 2;
	poses[1].layers[7].slant = slant / 6;
	poses[0].layers[7].slant = slant / 2;
	poses[0].layers[7].y = intent / 128;
	
	sprite->animation_time += dx / 512;
	
	/* do not assume 'char' is eight bits */
	sprite->animation_time &= 0xFF;
	
	if (sprite->animation_time < 128) {
		a = poses;
		b = poses + 1;
		i = sprite->animation_time;
	}
	else {
		a = poses + 1;
		b = poses;
		i = sprite->animation_time - 128;
	}
	
	y = sprite->landing_time * dx / 0x8000;
	
	a->layers[8].y = 128 + mimimi_sine[sprite->animation_time * 8];
	a->layers[8].y *= intent;
	a->layers[8].y /= 0x8000;
	a->layers[8].y += y;
	b->layers[8].y = a->layers[8].y;
	
	a->layers[9].y = 128 - mimimi_sine[sprite->animation_time * 8];
	a->layers[9].y *= intent;
	a->layers[9].y /= 0x8000;
	a->layers[9].y += y;
	b->layers[9].y = a->layers[9].y;
	
	poses[0].layers[0].y = 128 + mimimi_cosine[i * 16];
	poses[0].layers[0].y *= intent * (16 - sprite->landing_time);
	poses[0].layers[0].y /= 0x8000 * 16;
	poses[0].layers[0].y -= y;
	poses[1].layers[0].y = poses[0].layers[0].y;
	
	a->x = image->width / 2;
	a->y = image->height - 15;
	b->x = a->x;
	b->y = a->y;
	
	mimimi_interpolate(a, b, i, 128);
	mimimi_pose(image, sprite->model, a);
}

void mimimi_display_tick(struct mimimi_image *target, struct mimimi_sprite *sprite, struct mimimi_position *camera)
{
	static unsigned char colors[24 * 48];
	
	struct mimimi_physics *physics;
	struct mimimi_walk *walk;
	struct mimimi_image image;
	int x, y;
	int trail;
	
	physics = &sprite->physics;
	walk = &sprite->walk;
	
	image.width = 24;
	image.height = 48;
	image.colors = colors;
	
	for (y = 0 ; y < image.height ; y++) {
		for (x = 0 ; x < image.width ; x++) {
			colors[x + y * image.width] = 0;
		}
	}
	
	if (walk->direction != 0) sprite->direction = walk->direction;
	
	if (physics->airborne != 0) {
		sprite->landing_time = 16;
	}
	else {
		if (sprite->landing_time > 0) sprite->landing_time--;
	}
	
	if (physics->airborne == 0) mimimi_sprite_walking(&image, sprite);
	else mimimi_sprite_jumping(&image, sprite);
	
	trail = sprite->trail;
	sprite->trailing_y = mimimi_div(sprite->trailing_y * trail, 256);
	sprite->trailing_y += mimimi_div(sprite->position.y * (256 - trail), 256);
	sprite->trail = trail / 2;
	
	mimimi_camera_stamp(target, camera, sprite->position.x - image.width * 4, sprite->trailing_y - image.height * 8 + 32, 0, &image);
}

void mimimi_background_tick(struct mimimi_background *background, struct mimimi_image *image, struct mimimi_position *camera, struct mimimi_position *offset)
{
	mimimi_camera_stamp(image, camera, background->x + offset->x, background->y + offset->y, background->z, &background->image);
}
