/* license: AGPLv3 or later */
/* copyright 2024 zamfofex */

#include <stdlib.h>
#include <SDL2/SDL.h>
#include "../mimimi.h"

static struct SDL_Texture *mimimi_sdl2_texture(struct SDL_Renderer *renderer, struct mimimi_image *image)
{
	static unsigned char rch[] = {0x11, 0x44, 0x77, 0x99, 0xCC, 0xFF};
	static unsigned char gch[] = {0x11, 0x33, 0x55, 0x77, 0x99, 0xBB, 0xDD, 0xFF};
	static unsigned char bch[] = {0x22, 0x55, 0x88, 0xBB, 0xEE};
	
	int x, y;
	unsigned char color;
	struct SDL_Texture * texture;
	struct SDL_Surface * surface;
	Uint32 offset;
	Uint32 *pixel;
	Uint32 r, g, b, a;
	
	surface = SDL_CreateRGBSurface(0, image->width, image->height, 32, 0xFF0000, 0x00FF00, 0x0000FF, 0xFF000000L);
	
	for (y = 0 ; y < image->height ; y++) {
		
		for (x = 0 ; x < image->width ; x++) {
			
			offset = x * 4 + y * surface->pitch;
			pixel = (void *) ((char *) surface->pixels + offset);
			
			color = image->colors[x + y * image->width];
			if (color == 0) {
				*pixel = 0;
			}
			else {
				if (color < 0x10) {
					a = color * 0x11;
					*pixel = (0xFFL << 24) | (a << 16) | (a << 8) | (a << 0);
				}
				else {
					color -= 0x10;
					r = rch[(color / 40) % 6];
					g = gch[(color / 5) % 8];
					b = bch[(color / 1) % 5];
					a = 0xFF;
					*pixel = (a << 24) | (r << 16) | (g << 8) | (b << 0);
				}
			}
		}
	}
	
	texture = SDL_CreateTextureFromSurface(renderer, surface);
	
	SDL_FreeSurface(surface);
	
	return texture;
}

void mimimi_main(void (*tick)(void *data, struct mimimi_input *input), void *data)
{
	struct mimimi_input input;
	SDL_Window * window;
	SDL_Renderer * renderer;
	Uint32 tickno, now;
	SDL_Event event;
	unsigned char *keys;
	SDL_Texture * texture;
	
	atexit(&SDL_Quit);
	SDL_Init(SDL_INIT_VIDEO);
	
	input.image.width = 1024;
	input.image.height = 512;
	
	SDL_CreateWindowAndRenderer(input.image.width, input.image.height, SDL_WINDOW_RESIZABLE, &window, &renderer);
	SDL_SetWindowTitle(window, "Mirai's Miscellaneous Misadventures");
	
	tickno = 0;
	
	for (;;) {
		
		keys = (void *) SDL_GetKeyboardState(NULL);
		SDL_GetWindowSize(window, &input.image.width, &input.image.height);
		input.left = keys[SDL_SCANCODE_LEFT] || keys[SDL_SCANCODE_A];
		input.right = keys[SDL_SCANCODE_RIGHT] || keys[SDL_SCANCODE_D];
		(*tick)(data, &input);
		
		SDL_RenderSetLogicalSize(renderer, input.image.width, input.image.height);
		texture = mimimi_sdl2_texture(renderer, &input.image);
		SDL_RenderCopy(renderer, texture, NULL, NULL);
		SDL_DestroyTexture(texture);
		SDL_RenderPresent(renderer);
		
		now = SDL_GetTicks();
		do tickno++;
		while (tickno * 100 < now * 3) ;
		SDL_Delay(tickno * 100.0 / 3 - now);
		
		while (SDL_PollEvent(&event)) {
			if (event.type == SDL_QUIT) exit(0);
			if (event.type == SDL_KEYDOWN && event.key.keysym.sym == SDLK_q) exit(0);
		}
	}
}
